@@ -1,7 +1,7 @@
 """
 PySCeS - Python Simulator for Cellular Systems (http://pysces.sourceforge.net)
 
-Copyright (C) 2004-2013 B.G. Olivier, J.M. Rohwer, J.-H.S Hofmeyr all rights reserved,
+Copyright (C) 2004-2023 B.G. Olivier, J.M. Rohwer, J.-H.S Hofmeyr all rights reserved,
 
 Brett G. Olivier (bgoli@users.sourceforge.net)
 Triple-J Group for Molecular Cell Physiology
@@ -14,8 +14,12 @@ this distribution for specifics.
 NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.
 Brett G. Olivier
 """
+from __future__ import division, print_function
+from __future__ import absolute_import
+from __future__ import unicode_literals
 
 from pysces.version import __version__
+
 __doc__ = """
           PyscesScan
           ----------
@@ -27,14 +31,7 @@ import scipy
 from pysces.PyscesUtils import TimerBox
 
 __psyco_active__ = 0
-##  try:
-    ##  import psyco
-    ##  psyco.profile()
-    ##  __psyco_active__ = 1
-    ##  print 'PySCeS Module is now PsycoActive!'
-##  except:
-    ##  __psyco_active__ = 0
-##  __psyco_active__ = 0
+
 
 class Scanner(object):
     """
@@ -44,6 +41,7 @@ class Scanner(object):
     Steady-state results are always stored in self.SteadyStateResults while
     user output can be found in self.UserOutputResults - brett 2007.
     """
+
     genOn = True
     quietRun = False
     _MODE_ = 'state'
@@ -51,15 +49,17 @@ class Scanner(object):
     HAS_STATE_OUTPUT = True
     nan_on_bad_state = True
     MSG_PRINT_INTERVAL = 500
-    __AnalysisModes__ = ('state','elasticity','mca','stability','null')
+    __AnalysisModes__ = ('state', 'elasticity', 'mca', 'stability', 'null')
     invalid_state_list = None
     invalid_state_list_idx = None
 
     def __init__(self, mod):
         try:
             N = mod.nmatrix
-        except Exception, ex:
-            print '\nSCANNER: Please load a model <i.e. mod. doLoad()> before trying to use it.'
+        except Exception as ex:
+            print(
+                '\nSCANNER: Please load a model <i.e. mod. doLoad()> before trying to use it.'
+            )
 
         self.GenDict = {}
         self.GenOrder = []
@@ -70,7 +70,7 @@ class Scanner(object):
         self.UserOutputResults = []
         self.scanT = TimerBox()
 
-    def testInputParameter(self,name):
+    def testInputParameter(self, name):
         """
         This tests whether a str(name) is an attribute of the model
         """
@@ -85,9 +85,9 @@ class Scanner(object):
         parameters after a scan run.
         """
         for key in self.GenDict:
-            setattr(self.mod,key,self.GenDict[key][6])
+            setattr(self.mod, key, self.GenDict[key][6])
 
-    def addUserOutput(self,*kw):
+    def addUserOutput(self, *kw):
         """
         Add output parameters to the scanner as a collection of one or more
         string arguments ('O1','O2','O3', 'On'). These are evaluated at
@@ -95,30 +95,30 @@ class Scanner(object):
         array. The list of output is stored in self.UserOutputList.
         """
         output = [attr for attr in kw if type(attr) == str]
-        #print output
+        # print output
         self.HAS_USER_OUTPUT = True
         ModeList = list(self.__AnalysisModes__)
         MaxMode = 0
         for attr in output:
             if attr[:2] == 'ec' and self._MODE_ != 'null':
                 cM = ModeList.index('elasticity')
-                if  cM > MaxMode:
+                if cM > MaxMode:
                     self._MODE_ = 'elasticity'
                     MaxMode = cM
             elif attr[:2] == 'cc' and self._MODE_ != 'null':
                 cM = ModeList.index('mca')
-                if  cM > MaxMode:
+                if cM > MaxMode:
                     self._MODE_ = 'mca'
                     MaxMode = cM
             elif attr[:6] == 'lambda' and self._MODE_ != 'null':
                 cM = ModeList.index('stability')
-                if  cM > MaxMode:
+                if cM > MaxMode:
                     self._MODE_ = 'stability'
                     MaxMode = cM
-        print 'MaxMode', MaxMode
+        print('MaxMode', MaxMode)
         self.UserOutputList = output
 
-    def addScanParameter(self,name,start,end,points,log=False,slave=False):
+    def addScanParameter(self, name, start, end, points, log=False, follower=False):
         """
         Add a parameter to scan (an axis if you like) input is:
 
@@ -127,34 +127,44 @@ class Scanner(object):
         - float(end) = upper bound of scan
         - int(points) = number of points in scan range
         - bool(log) = Use a logarithmic (base10) range
-        - bool(slave) = Scan parameters can be masters i.e. an independent axis or a "slave" which moves synchronously with the previously defined parameter range.
+        - bool(follower) = Scan parameters can be leaders i.e. an independent axis or a "follower" which moves synchronously with the previously defined parameter range.
 
-        The first ScanParameter cannot be a slave.
+        The first ScanParameter cannot be a follower.
         """
         offset = 1
-        assert self.testInputParameter(name), '\nSCANNER: Model does not have an attribute \"%s\" \n' % name
-        assert name not in self.GenOrder, '\nSCANNER: This operation is currently not allowed\n'
+        assert self.testInputParameter(name), (
+            '\nSCANNER: Model does not have an attribute \"%s\" \n' % name
+        )
+        assert (
+            name not in self.GenOrder
+        ), '\nSCANNER: This operation is currently not allowed\n'
         if not len(self.GenOrder) == 0:
             for el in self.GenOrder:
-                if self.GenDict[el][4] == False:           # test if not slave
+                if self.GenDict[el][4] == False:  # test if not follower
                     offset = offset * self.GenDict[el][2]  # increment offset
-            if slave == True:                              
-                prevpar = self.GenOrder[-1]                # previous parameter, i.e. master
-                offset = self.GenDict[prevpar][5]          # don't increment for slave
+            if follower == True:
+                prevpar = self.GenOrder[-1]  # previous parameter, i.e. leader
+                offset = self.GenDict[prevpar][5]  # don't increment for follower
                 if points != self.GenDict[prevpar][2]:
-                    print 'SCANNER: Slave parameter needs to iterate over same number of\npoints as master...resetting points.'
+                    print(
+                        'SCANNER: Follower parameter needs to iterate over same number of\npoints as leader...resetting points.'
+                    )
                     points = self.GenDict[prevpar][2]
         else:
-            if slave == True:
-                slave = False
-                print 'SCANNER: Inner range cannot be a slave ... resetting to master'
-        self.GenDict.setdefault(name,(start,end,points,log,slave,offset,getattr(self.mod,name)))
-        setattr(self,name+'_test',self.stepGen(offset))
-        setattr(self,name,self.rangeGen(name,start,end,points,log))
+            if follower == True:
+                follower = False
+                print(
+                    'SCANNER: Inner range cannot be a follower ... resetting to leader'
+                )
+        self.GenDict.setdefault(
+            name, (start, end, points, log, follower, offset, getattr(self.mod, name))
+        )
+        setattr(self, name + '_test', self.stepGen(offset))
+        setattr(self, name, self.rangeGen(name, start, end, points, log))
         if name not in self.GenOrder:
             self.GenOrder.append(name)
 
-    def makeRange(self,start,end,points,log):
+    def makeRange(self, start, end, points, log):
         """
         Should be pretty self evident it defines a range:
 
@@ -164,12 +174,12 @@ class Scanner(object):
         - bool(log)
         """
         if log:
-            rng = scipy.logspace(scipy.log10(start),scipy.log10(end),points)
+            rng = scipy.logspace(scipy.log10(start), scipy.log10(end), points)
         else:
-            rng = scipy.linspace(start,end,points)
+            rng = scipy.linspace(start, end, points)
         return rng
 
-    def rangeGen(self,name,start,end,points,log):
+    def rangeGen(self, name, start, end, points, log):
         """
         This is where things get more interesting. This function creates
         a cycling generator which loops over a parameter range.
@@ -180,17 +190,17 @@ class Scanner(object):
         - *points*
         - *log* scale
         """
-        range = self.makeRange(start,end,points,log)
+        range = self.makeRange(start, end, points, log)
         cnt = 0
         while self.genOn:
-            if getattr(self,name+'_test').next():
-                if cnt >= len(range)-1:
+            if next(getattr(self, name + '_test')):
+                if cnt >= len(range) - 1:
                     cnt = 0
                 else:
                     cnt += 1
             yield range[cnt]
 
-    def stepGen(self,offset):
+    def stepGen(self, offset):
         """
         Another looping generator function. The idea here is to create
         a set of generators for the scan parameters. These generators then
@@ -211,19 +221,21 @@ class Scanner(object):
         """
         Fire all step generators
         """
-        return [getattr(self,genN+'_test').next() for genN in self.GenOrder]
+        return [next(getattr(self, genN + '_test')) for genN in self.GenOrder]
 
     def __nextValue__(self):
         """
         Fire all range generators
         """
-        return [getattr(self,genN).next() for genN in self.GenOrder]
+        return [next(getattr(self, genN)) for genN in self.GenOrder]
 
-    def setModValue(self,name,value):
+    def setModValue(self, name, value):
         """
         An easy one, assign value to name of the instantiated PySCeS model attribute
         """
-        assert hasattr(self.mod, name), '\nModel does not have an attribute: %s \n' % name
+        assert hasattr(self.mod, name), (
+            '\nModel does not have an attribute: %s \n' % name
+        )
         setattr(self.mod, name, float(value))
 
     def RunAgain(self):
@@ -234,13 +246,15 @@ class Scanner(object):
         """
         self.Run(ReRun=True)
 
-    def Run(self,ReRun=False):
+    def Run(self, ReRun=False):
         """
         Run the parameter scan
         """
         self.scanT.normal_timer('RUN')
         self._MODE_ = self._MODE_.lower()
-        assert self._MODE_ in self.__AnalysisModes__, '\nSCANNER: \"%s\" is not a valid analysis mode!' % self._MODE_
+        assert self._MODE_ in self.__AnalysisModes__, (
+            '\nSCANNER: \"%s\" is not a valid analysis mode!' % self._MODE_
+        )
         if self.quietRun:
             self.mod.SetQuiet()
         else:
@@ -254,21 +268,21 @@ class Scanner(object):
 
         if self.nan_on_bad_state:
             self.mod.__settings__["mode_state_nan_on_fail"] = True
-            #self.mod.mode_state_nan_on_fail = True
+            # self.mod.mode_state_nan_on_fail = True
         Tsteps = 1
         for gen in self.GenOrder:
-            if self.GenDict[gen][4] == False:      # don't increase Tsteps for slaves
+            if self.GenDict[gen][4] == False:  # don't increase Tsteps for follower
                 Tsteps *= self.GenDict[gen][2]
-        print self.scanT.RUN.next()
+        print(next(self.scanT.RUN))
         analysis_counter = 0
-        print 'SCANNER: Tsteps', Tsteps
+        print('SCANNER: Tsteps', Tsteps)
         pcntr = 0
         for step in range(Tsteps):
             pars = self.__nextValue__()
-            #if pars not in self.ScanSpace:
+            # if pars not in self.ScanSpace:
             self.ScanSpace.append(pars)
             for par in range(len(self.GenOrder)):
-                self.setModValue(self.GenOrder[par],pars[par])
+                self.setModValue(self.GenOrder[par], pars[par])
             analysis_counter += 1
             self.Analyze()
             if not self.mod.__StateOK__:
@@ -277,7 +291,7 @@ class Scanner(object):
             self.StoreData()
             pcntr += 1
             if pcntr >= self.MSG_PRINT_INTERVAL:
-                print '\t', analysis_counter, self.scanT.RUN.next()
+                print('\t', analysis_counter, next(self.scanT.RUN))
                 pcntr = 0
         self.ScanSpace = scipy.array(self.ScanSpace)
         self.SteadyStateResults = scipy.array(self.SteadyStateResults)
@@ -285,11 +299,11 @@ class Scanner(object):
         self.resetInputParameters()
         if self.nan_on_bad_state:
             self.mod.mode_state_nan_on_fail = False
-        print "\nSCANNER: %s states analysed\n" % analysis_counter
+        print("\nSCANNER: %s states analysed\n" % analysis_counter)
         if len(self.invalid_state_list) > 0:
-            print 'Bad steady states encountered at:\n'
-            print self.invalid_state_list
-            print self.invalid_state_list_idx
+            print('Bad steady states encountered at:\n')
+            print(self.invalid_state_list)
+            print(self.invalid_state_list_idx)
 
     def Analyze(self):
         """
@@ -313,9 +327,13 @@ class Scanner(object):
         Internal function which concatenates and stores the data generated by Analyze.
         """
         if self.HAS_STATE_OUTPUT and self._MODE_ != 'null':
-            self.SteadyStateResults.append(scipy.hstack((self.mod.state_species,self.mod.state_flux)))
+            self.SteadyStateResults.append(
+                scipy.hstack((self.mod.state_species, self.mod.state_flux))
+            )
         if self.HAS_USER_OUTPUT and self._MODE_ != 'null':
-            self.UserOutputResults.append(scipy.array([getattr(self.mod,res) for res in self.UserOutputList]))
+            self.UserOutputResults.append(
+                scipy.array([getattr(self.mod, res) for res in self.UserOutputList])
+            )
 
     def getOutput(self):
         """
@@ -339,14 +357,23 @@ class Scanner(object):
         labels = []
         if stst:
             if self.HAS_USER_OUTPUT:
-                output_array = scipy.hstack([self.ScanSpace, self.SteadyStateResults, self.UserOutputResults])
-                labels = self.GenOrder+list(self.mod.species)+list(self.mod.reactions)+self.UserOutputList
+                output_array = scipy.hstack(
+                    [self.ScanSpace, self.SteadyStateResults, self.UserOutputResults]
+                )
+                labels = (
+                    self.GenOrder
+                    + list(self.mod.species)
+                    + list(self.mod.reactions)
+                    + self.UserOutputList
+                )
             else:
                 output_array = scipy.hstack([self.ScanSpace, self.SteadyStateResults])
-                labels = self.GenOrder+list(self.mod.species)+list(self.mod.reactions)
+                labels = (
+                    self.GenOrder + list(self.mod.species) + list(self.mod.reactions)
+                )
         else:
             output_array = scipy.hstack([self.ScanSpace, self.UserOutputResults])
-            labels = self.GenOrder+self.UserOutputList
+            labels = self.GenOrder + self.UserOutputList
         if lbls:
             return output_array, labels
         else:
@@ -377,7 +404,7 @@ class PITCONScanUtils(object):
     res_flux = None
     res_eigen = None
 
-    def __init__(self,model):
+    def __init__(self, model):
         """
         Instantiate PITCONScanUtils with a loaded PySCeS model instance.
         """
@@ -390,7 +417,9 @@ class PITCONScanUtils(object):
         """
         self.user_output = [str(arg) for arg in args]
 
-    def runContinuation(self,parameter,low,high,density,par3d=None,logrange=True,runQuiet=True):
+    def runContinuation(
+        self, parameter, low, high, density, par3d=None, logrange=True, runQuiet=True
+    ):
         """
         Run the continuation using the following parameters:
 
@@ -402,7 +431,7 @@ class PITCONScanUtils(object):
         - density = int(the number of initial points)
         - par3d = float(extra 3d parameter to insert into the output array) this parameter is not set ONLY used in output
         - logrange = boolean [default = True], if True generate the result using logspace(log10(low), log10(high), density) otherwise use a linear range
-        - runQuiet = boolean [default = True], if True do not display intermediate results to screen, disable for debugging 
+        - runQuiet = boolean [default = True], if True do not display intermediate results to screen, disable for debugging
 
         After running the continuation the results are stored in numpy arrays
 
@@ -411,39 +440,44 @@ class PITCONScanUtils(object):
         - mod.res_flux = steady-state flux values
 
         """
-        
+
         self.pitcon_scan_density = density
         self.pitcon_scan_parameter = parameter
         self.pitcon_scan_parameter_3d = par3d
         if logrange:
             self.pitcon_range_low = scipy.log10(low)
             self.pitcon_range_high = scipy.log10(high)
-            self.model.pitcon_par_space = scipy.logspace(self.pitcon_range_low,self.pitcon_range_high,self.pitcon_scan_density)
+            self.model.pitcon_par_space = scipy.logspace(
+                self.pitcon_range_low, self.pitcon_range_high, self.pitcon_scan_density
+            )
         else:
             self.pitcon_range_low = low
             self.pitcon_range_high = high
-            self.model.pitcon_par_space = scipy.linspace(self.pitcon_range_low,self.pitcon_range_high,self.pitcon_scan_density)            
+            self.model.pitcon_par_space = scipy.linspace(
+                self.pitcon_range_low, self.pitcon_range_high, self.pitcon_scan_density
+            )
 
         self.model.pitcon_flux_gen = 1
         if runQuiet:
             self.model.SetQuiet()
         else:
             self.model.SetLoud()
-            
+
         if self.pitcon_scan_parameter_3d != None:
-            self.pitcon_res = self.model.PITCON(self.pitcon_scan_parameter, self.pitcon_scan_parameter_3d)
-            self.res_idx = self.pitcon_res[:,:2]
-            self.res_metab = self.pitcon_res[:,2:len(self.model.species)+2:]
-            self.res_flux = self.pitcon_res[:,len(self.model.species)+2:]
+            self.pitcon_res = self.model.PITCON(
+                self.pitcon_scan_parameter, self.pitcon_scan_parameter_3d
+            )
+            self.res_idx = self.pitcon_res[:, :2]
+            self.res_metab = self.pitcon_res[:, 2 : len(self.model.species) + 2 :]
+            self.res_flux = self.pitcon_res[:, len(self.model.species) + 2 :]
         else:
             self.pitcon_res = self.model.PITCON(self.pitcon_scan_parameter)
-            self.res_idx = self.pitcon_res[:,0]
-            self.res_idx = self.res_idx.reshape(self.res_idx.shape[0],1)
-            self.res_metab = self.pitcon_res[:,1:len(self.model.species)+1]
-            self.res_flux = self.pitcon_res[:,len(self.model.species)+1:]
-
-        print '\n\tContinuation complete\n'
+            self.res_idx = self.pitcon_res[:, 0]
+            self.res_idx = self.res_idx.reshape(self.res_idx.shape[0], 1)
+            self.res_metab = self.pitcon_res[:, 1 : len(self.model.species) + 1]
+            self.res_flux = self.pitcon_res[:, len(self.model.species) + 1 :]
 
+        print('\n\tContinuation complete\n')
 
     def analyseData(self, analysis='elas'):
         """
@@ -466,20 +500,20 @@ class PITCONScanUtils(object):
         self.res_user = []
 
         step_cntr = 0
-        self.timer.step_timer('st1',self.res_idx.shape[0])
+        self.timer.step_timer('st1', self.res_idx.shape[0])
         for row in range(self.res_idx.shape[0]):
             self.timer.st1_cstep += 1
             step_cntr += 1
             if self.timer.st1_cstep == 1 or step_cntr == 200:
-                print '\n', self.timer.st1.next()
+                print('\n', next(self.timer.st1))
                 step_cntr = 0
 
-            PitParVal = self.res_idx[row,-1]
-            setattr(self.model,self.pitcon_scan_parameter,PitParVal)
+            PitParVal = self.res_idx[row, -1]
+            setattr(self.model, self.pitcon_scan_parameter, PitParVal)
 
-            self.model.state_species = self.res_metab[row,:]
-            self.model.state_flux = self.res_flux[row,:]
-            self.model.SetStateSymb(self.model.state_flux,self.model.state_species)
+            self.model.state_species = self.res_metab[row, :]
+            self.model.state_flux = self.res_flux[row, :]
+            self.model.SetStateSymb(self.model.state_flux, self.model.state_species)
             self.model.Forcing_Function()
 
             if analysis in ['elasv', 'elas', 'mca', 'resp', 'eigen', 'all']:
@@ -498,12 +532,11 @@ class PITCONScanUtils(object):
                 for attr in self.user_output:
                     try:
                         state_out.append(getattr(self.model, attr))
-                    except Exception, ex:
-                        print ex
+                    except Exception as ex:
+                        print(ex)
                 self.res_user.append(state_out)
         self.res_user = scipy.array(self.res_user)
-        print '\n\t%s evaluation complete\n' % analysis
-
+        print('\n\t%s evaluation complete\n' % analysis)
 
     def getArrayListAsArray(self, array_list):
         """
@@ -515,7 +548,3 @@ class PITCONScanUtils(object):
         elif len(array_list) == 1:
             output = array_list[0]
         return output
-
-if __psyco_active__:
-    psyco.bind(Scanner)
-    psyco.bind(PITCONScanUtils)
